@@ -14,6 +14,7 @@ use IO::File qw(!/^SEEK/);
 use POSIX;
 use Fcntl qw(:flock);
 use File::Basename;
+use File::Path;
 use HTML::Mason::Config;
 use HTML::Mason::Tools qw(date_delta_to_secs);
 use MLDBM ($HTML::Mason::Config{mldbm_use_db}, $HTML::Mason::Config{mldbm_serializer});
@@ -40,8 +41,23 @@ sub access_data_cache
 	$lockdir .= "locks";
 	mkpath($lockdir,0,0755) if (!-d $lockdir);
 	my $lockfile = "$lockdir/$base.lock";
-	my $lockfh = new IO::File ">>$lockfile"
-	   or die "cache: cannot open lockfile '$lockfile' for locking\n";
+
+	# Open file in correct mode for lock type (Tom Hughes)
+	my $lockfh;
+	if ($lockargs & LOCK_EX) {
+	    $lockfh = new IO::File ">>$lockfile"
+		or die "cache: cannot open lockfile '$lockfile' for exclusive lock: $!";
+	} elsif ($lockargs & LOCK_SH) {
+	    $lockfh = new IO::File "<$lockfile";
+	    if (!$lockfh && !-e $lockfile) {
+		$lockfh = new IO::File ">$lockfile";
+		$lockfh->close;
+		$lockfh = new IO::File "<$lockfile";
+	    }
+	    die "cache: cannot open lockfile '$lockfile' for shared lock: $!" if !$lockfh;
+	} else {
+	    die "unknown lock mode: $lockargs";
+	}
 	return (flock($lockfh, $lockargs)) ? $lockfh : undef;
     };
     
@@ -290,3 +306,16 @@ sub access_data_cache
     }
 }
 
+#
+# Returns 1 if the exclusive, non-blocking lock was obtained,
+# undef otherwise. Left here for content management
+# backwards compatibility!
+#
+sub get_lock {
+    my $fh = shift;
+
+    my $LOCK_EX = 2;
+    my $LOCK_NB = 4;
+    my $LOCK_UN = 8;
+    return flock $fh, $LOCK_EX|$LOCK_NB;
+}
